#include <math.h>
#include <graphics.h>
#include <stdlib.h>

extern "C"
{
	#include "detect.h"
	#include "smix.h"
}

#include "Aa3Anim.H"
#include "Aa3Math.H"
#include "Aa3Draw.H"
#include "Aa3Snds.H"
#include "Aa3Const.H"

const float pi_6 = 3.14159/6;
const float pi_180 = 3.14159/180;
const float sqrt2 = sqrt(2);


// extern of time
extern unsigned long time_elapsed, score;
extern char dead, dashboard_change;
static char locked_on_switched;

extern char SFX;                // in aa3snds.cpp
extern SOUND *Sound[NUMSOUNDS];


void AnimateY(YShip &Y, EShipPtr E, EShipPtr &locked_on)
{
	if (locked_on == NULL)
   	if (E->next != E)
      {
      	locked_on = E->next;
         locked_on_switched = 1;
      }

   if (Y.lastmov + MOVE_DELAY <= time_elapsed)
   	if (Y.bearing.mag > 0)
   	{
      	EraseYShip(Y.x, Y.y);
   		Y.x += Y.bearing.mag * cos(Y.bearing.theta);
     	 	Y.y += Y.bearing.mag * sin(Y.bearing.theta);
         if (Y.x < 0) Y.x += 640;
         if (Y.x > 639) Y.x -= 640;
         if (Y.y < 0) Y.y += 440;
         if (Y.y > 439) Y.y -= 440;
         Y.lastmov = time_elapsed;
         DrawYShip(Y.x, Y.y, Y.aim);
   	}

   if (Y.lastrecharge + RECHARGE_DELAY <= time_elapsed)
   {
    	if (Y.power < 100)
      {
      	Y.power += POWER_RECHARGE_AMT;
         if (Y.power > 100)
         {
         	Y.shields += floor((Y.power-100.0)*SHIELD_RECHARGE_AMT/POWER_RECHARGE_AMT);
            if (Y.shields > 100) Y.shields = 100;
            Y.power = 100;
         }
         dashboard_change = 1;
      }
      else if (Y.shields < 100)
      {
      	Y.shields += SHIELD_RECHARGE_AMT;
         dashboard_change = 1;
      }
      Y.lastrecharge = time_elapsed;
   }

   float locked_on_angle = get_theta(Y.x, Y.y, locked_on->x, locked_on->y);  // raidans
   EShipPtr cur, del;
   cur = E->next;
   while (cur != E && !dead)
   {
   	if (abs(cur->x - Y.x) < Y_SIDE_LENGTH+ENEMY_RADIUS && abs(cur->y - Y.y) < Y_SIDE_LENGTH+ENEMY_RADIUS)
      {
      	Y.shields -= CRASH_DAMAGE;
         score += ALIEN_POINTS*cur->ship_type;
         del = cur;
         cur = cur->prev;
         cur->next = del->next;
         cur->next->prev = cur;
         delete del;
         Explode(Y.x, Y.y);
         if (Y.shields < 0)
         	dead = 1;
         else
         {
         	dashboard_change = 1;
            DrawYShip(Y.x, Y.y, Y.aim);
         }
      }
      else if (delta_t(Y.aim, locked_on_angle) > pi_6 && delta_t(Y.aim, get_theta(Y.x, Y.y, cur->x, cur->y)) <= pi_6)
      {
      	locked_on = cur;
         locked_on_switched = 1;
      }
      cur = cur->next;
   }
}


void AnimateYP(YProjsPtr YP, EShipPtr E, EShipPtr &locked_on)
{
	YProjsPtr y_cur, y_del;
   EShipPtr e_cur, e_del;
   char exploded;

   y_cur = YP->next;
   while (y_cur != YP)
   {
   	exploded = 0;
   	e_cur = E->next;
      while (e_cur != E && !exploded)
      {
      	if (abs(y_cur->x - e_cur->x) < BULLET_RADIUS+ENEMY_RADIUS+5 &&
             abs(y_cur->y - e_cur->y) < BULLET_RADIUS+ENEMY_RADIUS+5 &&
             y_cur->answer == e_cur->answer)
         {
         	if (locked_on == e_cur)
            {
            	locked_on = NULL;
               locked_on_switched = 1;
            }
         	Explode(y_cur->x, y_cur->y);
            score += ALIEN_POINTS*e_cur->ship_type;
            e_del = e_cur;
            y_del = y_cur;
            e_cur = e_cur->prev;
            y_cur = y_cur->prev;
            e_cur->next = e_del->next;
            y_cur->next = y_del->next;
            e_cur->next->prev = e_cur;
            y_cur->next->prev = y_cur;
            delete e_del;
            delete y_del;
            exploded = 1;
         }
         e_cur = e_cur->next;
      }

      if (!exploded && y_cur->lastmove + Y_BULLET_MOVE_DELAY <= time_elapsed)
	      if (y_cur->bearing.mag > 0)
	      {
         	EraseYP(y_cur->x, y_cur->y);
	      	y_cur->x += y_cur->bearing.mag * cos(y_cur->bearing.theta) * Y_BULLET_MOVE_COEFF;
            y_cur->y += y_cur->bearing.mag * sin(y_cur->bearing.theta) * Y_BULLET_MOVE_COEFF;
            if (y_cur->x < 0) y_cur->x += 640;
	         if (y_cur->x > 639) y_cur->x -= 640;
	         if (y_cur->y < 0) y_cur->y += 440;
	         if (y_cur->y > 439) y_cur->y -= 440;
            y_cur->range -= Y_BULLET_MOVE_COEFF;
            y_cur->lastmove = time_elapsed;

            if (y_cur->range <= 0)
            {
               y_del = y_cur;
               y_cur = y_cur->prev;
               y_cur->next = y_del->next;
               y_cur->next->prev = y_cur;
               delete y_del;
            }
            else DrawYP(y_cur->x, y_cur->y, y_cur->answer);
         }

      y_cur = y_cur->next;
   }
}


void AnimateEP(EProjsPtr EP, YShip &Y)
{
	EProjsPtr e_cur, e_del;

   e_cur = EP->next;
   while (e_cur != EP && !dead)
   {
   	if (abs(e_cur->x - Y.x) < E_PROJ_RADIUS+Y_SIDE_LENGTH &&
      	 abs(e_cur->y - Y.y) < E_PROJ_RADIUS+Y_SIDE_LENGTH)
      {
      	if (e_cur->type == 0)
      		Y.shields -= 10;
         if (e_cur->type == 1)
         	Y.shields -= 25;
         if (Y.shields < 0)
         {
            dead = 1;
            Explode(Y.x, Y.y);
         }
         else
         {
         	SmEx(e_cur->x, e_cur->y);
         	HitYou();
            dashboard_change = 1;
            DrawYShip(Y.x, Y.y, Y.aim);
            e_del = e_cur;
            e_cur = e_cur->prev;
            e_cur->next = e_del->next;
            e_cur->next->prev = e_cur;
            delete e_del;
         }
      }
      else // if you have not been hit
      {
      	if (e_cur->lastmove + E_PROJ_MOVE_DELAY <= time_elapsed)
		      if (e_cur->bearing.mag > 0)
		      {
  		       	EraseEP(e_cur->x, e_cur->y);
		      	e_cur->x += e_cur->bearing.mag * cos(e_cur->bearing.theta) * E_PROJ_MOVE_COEFF;
	            e_cur->y += e_cur->bearing.mag * sin(e_cur->bearing.theta) * E_PROJ_MOVE_COEFF;
	            if (e_cur->x < 0) e_cur->x += 640;
		         if (e_cur->x > 639) e_cur->x -= 640;
		         if (e_cur->y < 0) e_cur->y += 440;
		         if (e_cur->y > 439) e_cur->y -= 440;
	            e_cur->range -= E_PROJ_MOVE_COEFF;
	            e_cur->lastmove = time_elapsed;

	            if (e_cur->range <= 0)
	            {
               	if (e_cur->type)  // IF FLAK
                  {
                  	FlakExplode(e_cur->x, e_cur->y, Y);
                     DrawYShip(Y.x, Y.y, Y.aim);
                  }
	               e_del = e_cur;
	               e_cur = e_cur->prev;
	               e_cur->next = e_del->next;
	               e_cur->next->prev = e_cur;
	               delete e_del;
	            }
	            else DrawEP(e_cur->x, e_cur->y, e_cur->type);
   	      }
      }
      e_cur = e_cur->next;
   }
}

void AnimateE(EShipPtr E, EShipPtr &locked_on, EProjsPtr EP, YShip &Y, ETypePtr ET)
{
	if (locked_on == NULL && E->next != E)
   {
   	locked_on = E->next;
      locked_on_switched = 1;
   }

   EShipPtr cur, del;
	cur = E->next;
   while (cur != E)
   {
   	if (cur->lastturn + E_TURN_WAIT <= time_elapsed)
	   {
	   	cur->bearing.mag *= (random(20)+1)/10.0;
	      cur->bearing.theta += (random(11)-5)*pi_180;
	      cur->lastturn = time_elapsed;
	   }

	   if (cur->lastmov + E_MOVE_WAIT <= time_elapsed)
	   {
	   	EraseE(cur->x, cur->y);
	   	cur->x += cur->bearing.mag * cos(cur->bearing.theta) * E_MOVE_COEFF;
	      cur->y += cur->bearing.mag * sin(cur->bearing.theta) * E_MOVE_COEFF;
	      cur->lastmov = time_elapsed;
         if (cur->x < -ENEMY_RADIUS-1 ||
         	 cur->x > 640+ENEMY_RADIUS ||
             cur->y < -ENEMY_RADIUS-1 ||
             cur->y > 440+ENEMY_RADIUS)
         {
         	if (locked_on == cur)
            {
            	locked_on = NULL;
               locked_on_switched = 1;
            }
         	del = cur;
            cur = cur->prev;
            cur->next = del->next;
            cur->next->prev = cur;
            delete del;
            continue;
         }
	      DrawE(cur->x, cur->y, cur->ship_type, locked_on==cur);
	   }

	   if (cur->lastshoot + E_FIRE_WAIT <= time_elapsed)
	   {
         if (SFX)
         	if (cur->ship_type == 4)
            	start_sound(Sound[4], 4, 0);
            else
            	start_sound(Sound[3], 3, 0);

	   	EProjsPtr ins;

	      ins = new EProjs;
	      if (cur->ship_type == 4)  // flak ship
	   	{
	      	ins->type = 1;
	         ins->range = distance(cur->x, cur->y, Y.x, Y.y) + random(11)-5;
	      }
	      else
	      {
	      	ins->type = 0;
	         ins->range = E_PROJ_RANGE*cur->ship_type;
	      }
			ins->bearing.mag = E_PROJ_FIRE_MAG;
	      ins->bearing.theta = get_theta(cur->x, cur->y, Y.x, Y.y);
	      ins->x = cur->x + (ENEMY_RADIUS*sqrt2)*cos(ins->bearing.theta);
	      ins->y = cur->y + (ENEMY_RADIUS*sqrt2)*sin(ins->bearing.theta);
	      ins->lastmove = 0;
	      ins->prev = EP;
	      ins->next = EP->next;
	      EP->next = ins;
	      ins->next->prev = ins;

         cur->lastshoot = time_elapsed;
   	}

	   cur = cur->next;
	}

   if (locked_on_switched)
   {
   	locked_on_switched = 0;
      if (locked_on == NULL && E->next != E) locked_on = E->next;
      int i;
      if (locked_on != NULL)
      {
      	ETypePtr trav = ET;
      	for (i = 0; i < locked_on->question; i++) trav = trav->next;
      	UpdateQuestion(trav->question);
      }
      else
      {
      	char *stuff = NULL;
         UpdateQuestion(stuff);
      }
   }
}


void AnimateS(int_coords S[], YShip &Y)
{
	for (int i = 0; i < NUM_STARS; i++) 
   {
		if ( !(abs(S[i].x - Y.x) <= Y_SIDE_LENGTH+1 &&
          abs(S[i].y - Y.y) <= Y_SIDE_LENGTH+1) )
      	putpixel(floor(S[i].x), floor(S[i].y), random(16));
	}
}
